/* 
 * Copyright (C) 2008 Laird Breyer
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Author:   Laird Breyer <laird@lbreyer.com>
 */

#include "xtwin.h"
#include "util.h"
#include "bittwiddler.h"

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <signal.h> /* defines SIGWINCH equals 28 */

bool_t create_xtwin(xtwin_t *w, int maxwidth, int maxheight) {
  XPixmapFormatValues *pixmap_formats;
  XSizeHints hints;
  int count, i;

  if( w ) {
    w->dpy = XOpenDisplay(NULL);
    w->vis = DefaultVisual(w->dpy, 0);
    w->scr = DefaultScreen(w->dpy);
    w->depth = DefaultDepth(w->dpy, w->scr);

    pixmap_formats = XListPixmapFormats(w->dpy, &count);
    for(i = 0; i < count; i++) {
      if( pixmap_formats[i].depth == w->depth ) {
	w->bytespp = pixmap_formats[i].bits_per_pixel/8;
	w->scanline_pad = pixmap_formats[i].scanline_pad;
      }
    }
    XFree(pixmap_formats);

    w->xsize = 320;
    w->ysize = 200;
    w->maxwidth = maxwidth;
    w->maxheight = maxheight;
    w->win = XCreateWindow(w->dpy,
			   DefaultRootWindow(w->dpy),
			   100, 100,
			   w->xsize, w->ysize,
			   0,
			   CopyFromParent,
			   CopyFromParent,
			   w->vis,
			   0, NULL);
    XStoreName(w->dpy, w->win, "tessboxes");

    hints.flags = PPosition|PMinSize|PMaxSize;
    hints.x = w->xsize;
    hints.y = w->ysize;
    hints.min_width = w->xsize;
    hints.min_height = w->ysize;
    hints.max_width = w->maxwidth;
    hints.max_height = w->maxheight;
    XSetWMNormalHints(w->dpy, w->win, &hints);

    XMapWindow(w->dpy, w->win);
    XSelectInput(w->dpy, w->win, 
		 StructureNotifyMask|PropertyChangeMask|ExposureMask);

    sync_xtwin(w);

    w->xim = NULL;
    w->sbuf = NULL;
    alloc_image_xtwin(w);

    return TRUE;
  }
  return FALSE;
}


bool_t free_xtwin(xtwin_t *w) {
  if( w ) {
    if( w->xim ) { XDestroyImage(w->xim); }
    if( w->win ) { XDestroyWindow(w->dpy, w->win); }
    if( w->dpy ) { XCloseDisplay(w->dpy); }
    w->dpy = NULL;
    w->win = 0;
    w->vis = NULL;
    w->xim = NULL;
    w->sbuf = NULL;
  }
  return FALSE;
}

/* synchronize data structures with window geometry in
   case user changed them */
bool_t sync_xtwin(xtwin_t *w) {
  XWindowAttributes attr;
  XGetWindowAttributes(w->dpy, w->win, &attr);
  if( (w->xsize != attr.width) ||
      (w->ysize != attr.height) ) {
    w->xsize = attr.width;
    w->ysize = attr.height;

    return TRUE;
  }
  return FALSE;
}

bool_t alloc_image_xtwin(xtwin_t *w) {

  if( w->xim ) { XDestroyImage(w->xim); }

  w->sbuf = 
    (char *)malloc(w->xsize * w->ysize * w->bytespp * sizeof(char));
  w->xim = XCreateImage(w->dpy, w->vis, w->depth,
			ZPixmap, 0, 
			w->sbuf,
			w->xsize, w->ysize,
			w->scanline_pad,
			0);


  return TRUE;
}

xtwin_event_t next_event_xtwin(xtwin_t *w) {
  XEvent event;

  XFlush(w->dpy); 

  if( XPending(w->dpy) ) {
    XNextEvent(w->dpy, &event);
    switch(event.type) {
    case MapNotify:
    case Expose:
      return READY_EVENT;
    case PropertyNotify:
      if( sync_xtwin(w) ) {
	alloc_image_xtwin(w);
	return READY_EVENT;
      }
      break;
    default:
      break;
    }
    return NULL_EVENT;
  }
  return NO_EVENTS;
}

bool_t redraw_xtwin(xtwin_t *w, imagebuffer_t *ib, box_t *box) {
  char *c;
  int x, y, z;
  int bcx, bcy;
  int xmin, xmax, ymin, ymax;
  int bxmin, bxmax, bymin, bymax;
  imagepoint_t *point;
  byte_t red, green, blue;
  bool_t h;
  
  if( w && ib && box ) {
    
    bcx = (box->left + box->right) / 2;
    bcy = ib->height - (box->top + box->bottom) / 2;
    
    xmin = bcx - w->xsize/2;
    xmin = (xmin >= 0) ? xmin : 0;

    xmax = xmin + w->xsize;
    xmax = (xmax <= ib->width) ? xmax : ib->width;

    ymin = bcy - w->ysize/2;
    ymin = (ymin >= 0) ? ymin : 0;

    ymax = ymin + w->ysize;
    ymax = (ymax <= ib->height) ? ymax : ib->height;

    bxmin = bcx - (box->right - box->left) / 2; 
    bxmax = bxmin + (box->right - box->left);
    bymin = bcy - (box->bottom - box->top) / 2; 
    bymax = bymin + (box->bottom - box->top) - 1;

    c = w->sbuf;
    for(y = ymin; y < ymax; y++) {
      point = ib->image + y * ib->width + xmin;
      for(x = xmin; x < xmax; x++) {
	render_pixel_imagebuffer(point, 0, &red, &green, &blue);
	point++;

	/* highlight flag */
	h = (x >= bxmin) && (x <= bxmax) && (y >= bymin) && (y < bymax);

	/* quick and dirty hack to avoid dealing with X11 visuals,
	   give each byte same value */
	for(z = 0; z < w->bytespp; z++) {
	  *c++ = h ? red + 64: red;
	}

      }
    }

    XPutImage(w->dpy, w->win, DefaultGC(w->dpy, 0), w->xim,
	      0, 0, 0, 0, w->xsize, w->ysize);
    return TRUE;
  }
  return FALSE;
}
